// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#include<iostream>
#include<vector>
#include<string>
#include"param.h"
#include <sstream>
#include <fstream>
#include <cctype>
#include <iomanip>
#include <cstring>
#include "errstr.h"
#include "Dal.h"
#include<cmath>
#include <f90cxx.h>


#include"omphotom.h"
#include "OmSourceParameters.h"



/***********************************************************************
*
***********************************************************************/
class Parameters
{
public:
    int sourceNo;
    int type;
    double radius;
    double innerA;
    double outerA;
    double axialRatio;
    double theta;
    double background;


    /**************************************
    * For PSF fitting
    ***************************************/
    int noTerms;
    int fitNo[10];    

    int weightingMethod;



    Parameters() : 
        sourceNo(0), type(0), radius(0.0), innerA(0.0), 
             outerA(0.0), axialRatio(0.0), theta(0.0), background(0.0), noTerms(1), weightingMethod(1) {};  

    Parameters(int no, int t, double r, double a, double b, double ax, double th, double bck, int method) : 
        sourceNo(no), type(t), radius(r), innerA(a), outerA(b), axialRatio(ax), theta(th), background(bck), 
           noTerms(1) {fitNo[0] = no, weightingMethod = method;};  


    Parameters(int no, int t, double r, double a, double b, double ax, double th, double bck, int nt, int method) : 
        sourceNo(no), type(t), radius(r), innerA(a), outerA(b), axialRatio(ax), theta(th), background(bck), noTerms(nt) 
        { weightingMethod = method; };  
    


    
   ~Parameters() {};
};


bool getNextLine(ifstream &input, string &buffer, int length, bool warn = true);


void getWindowData(string file, int &binax1);
void process(vector<Parameters> &parameters);
int length(string & s);

/*************************************************************************
*
**************************************************************************/
void readParameterFile(vector<string> &imageFile,
                       vector<string> &backgroundImageFile, 
                       vector<string> &oldSrcListFile, 
                       vector<string> &newSrcListFile,
                       vector< vector<Parameters> > &paramaters);





int getNoRows(string &file);



/****************************************************
*
****************************************************/
int getInteger(stringstream &ss, string &buffer);
double getDouble(stringstream &ss, string &buffer);

int getInteger(string &buffer);
double getDouble(string &buffer);

/**************************************************
*
**************************************************/

string getString(string &buffer, int minLength);

string removeBlanks( string & s);



/**********************************************************
*
***********************************************************/
bool checkFiles(string &imageSet, string &backgroundSet, string &oldSrcListSet);

bool checkFiles(string &imageSet, string &oldSrcListSet);


/*****************************************************************
* FORTRAN interface- so can use existing fortran photometry module
*****************************************************************/
F90MODPROCPROTO(void, omphotom_f1, initialise_data)(int *nSources);

F90MODPROCPROTO(void, omphotom_f1, recompute_photometry)(int *numberToFit, int *no, 
                int *, double *, double *, double *, double*, double*, double*);

F90MODPROCPROTO(void, omphotom_f1, getoldsource)(int *no, OmSourceParameter *);
F90MODPROCPROTO(void, omphotom_f1, getnewsource)(int *no, OmSourceParameter *);
F90MODPROCPROTO(void, omphotom_f1, finish)();



bool logFile;
ofstream output;
/**************************************************************************
* The main program
*
**************************************************************************/
void omphotom()
{


    
    /*************************************************************************
    * Get the parameters
    **************************************************************************/


    string file = stringParameter("logfile");
    if (file.length() > 1)
    {
        logFile = true;
        output.open( file.c_str() );
    }
    else
        logFile = false;     


    file = stringParameter("paramfile");

    if(file.length() > 2)
    {

        /**********************************************
	* Input from a parameter file
	*********************************************/

        vector<string> imageFileList; 
        vector<string> backgroundImageFileList; 

        vector<string> oldSrcListFileList; 
        vector<string> newSrcListFileList;
        vector< vector<Parameters> > parameterList;                         
        


        readParameterFile( imageFileList,
                           backgroundImageFileList,
                           oldSrcListFileList, 
                           newSrcListFileList,
                           parameterList);
   
        /**********************************************************
	* Process image files one by one
        ***********************************************************/
        for(unsigned i = 0; i < imageFileList.size(); i++)
        {
	
	    if(logFile)
	    {
              
	      output << "Image file=" << imageFileList[i] << endl;
              output << "Background image file=" << backgroundImageFileList[i] << endl;
              output << "Old srclist file=" << oldSrcListFileList[i] << endl;
              output << "New srclist file=" << newSrcListFileList[i] << endl;
              output << endl;


            }
	    setParameter("imageset", imageFileList[i]);
	    setParameter("backgroundimageset", backgroundImageFileList[i]);
            setParameter("oldsrclistset", oldSrcListFileList[i]);
            setParameter("newsrclistset", newSrcListFileList[i]);
	    process( parameterList[i] );
            if (logFile)
	    {
                if (i < (imageFileList.size() - 1))
		  output << endl << endl << endl;
	    }


        }

    }
    else // Process single image file and source list
    {


        string file = stringParameter("oldsrclistset");
        int noOfRows = getNoRows( file );


 

        // The list of source numbers
        vector<int> sourceNo = intListParameter("sourceno");


        if(sourceNo.size() == 0 )
        {

            Sas::warning("No sources to reprocess!", "Please start again!");
            exit(0);

        }


        // The list of source numbers
        vector<int> photoType = intListParameter("phototype");
  


        if(sourceNo.size() != photoType.size() )
        {

            Sas::fatal("No of parameters of sourceno not equal to no of parameters of phototype", "Please try again!");

        }







     
        // The list of apertures
        vector<double> radius = realListParameter("aperturer");



        // The list of inner background annulli
        vector<double> innerA = realListParameter("innera");

        // The list of outer background annulli

        vector<double> outerA = realListParameter("outera");

        // The list of axial ratios
        vector<double> axialRatio = realListParameter("axialratio");

        // The list of aperture position angles
        vector<double> pa = realListParameter("pa");


        // The list of background counts per pixel to be used for a given source
        vector<double> background = realListParameter("background");
     
        vector<Parameters> parameters;
  

        // Need to find out if any photometry is extended
        bool extended = false;

    
        bool needBackground = false;   
        bool needAperture = 0;
        bool needBackgroundAnnulus = false;



        int weightingMethod =  intParameter("weighting");
        for(unsigned i = 0; i < photoType.size(); i++)
	{
	 

            if(sourceNo[i] < 0 || sourceNo[i] > noOfRows)
	    {
                Sas::fatal("No such source in file", stringParameter("oldsrclistset"));
     
            }

            switch(photoType[i])     
	    {
                   
	        case -2: // Fixed background
		    needBackground = true;
                    extended = true;
                    needAperture = true;
                   

                    break;
                
                case -1: 
                    needAperture = true;
                    needBackgroundAnnulus = true;
                    extended = true;
                    break;
             
	         case 0: // Normal aperture photometry  

                    needAperture = true;
                    needBackgroundAnnulus = true;
                  
                    break;
           
	         case 1:      // Aperture photometry with fixed background 
                    needAperture = true;              
	            needBackground = true;;


		
                    break;

	         case 2:     // PSF fit - background from anulus  
                                 
	            needBackgroundAnnulus = true;;
                    break;

	         case 4: case 6:  // PSF fit to several sources with fixed background
                    needBackground = true;;
                    break;


                   
	        default:      
		    break;  
                   
            }
        }
        /*************************************************************************
	* Some basic checks on parameters
        *************************************************************************/

        if(needAperture)
	{



            if(sourceNo.size() != radius.size() )
            {

                Sas::fatal("No of parameters of sourceno not equal to no of parameters of radius", "Please try again!");

            }
        }
      
   

        if(needBackground)
	{


            if(sourceNo.size() != background.size() )
            {

                Sas::fatal("No of parameters of sourceno not equal to no of parameters of background", "Please try again!");

            }
        }

      
        if(needBackgroundAnnulus)
	{

            if(sourceNo.size() != innerA.size() )
                {

                    Sas::fatal("No of parameters of sourceno not equal to no of parameters of innerA", "Please try again!");

                }

                if(sourceNo.size() != outerA.size() )
                {

                Sas::fatal("No of parameters of sourceno not equal to no of parameters of outerA", "Please try again!");

                }

	}


             
	if(extended)
	{          
            if(sourceNo.size() != axialRatio.size() )
            {

                Sas::fatal("No of parameters of sourceno not equal to no of parameters of axialratio", "Please try again!");

            }
            if(sourceNo.size() != pa.size() )
            {

                Sas::fatal("No of parameters of sourceno not equal to no of parameters of pa", "Please try again!");

            }
              

            if( needBackground &&  !needBackgroundAnnulus)  // Using fixed background for all sources
	    {  
                for(int i = 0; i < parameterCount("sourceno"); i++)
	        {
	            parameters.push_back(Parameters(sourceNo[i], photoType[i], 
                        radius[i], 0.0, 0.0, axialRatio[i], pa[i], background[i], weightingMethod) );
                }
            }
            else if( needBackground )  // Using fixed background for some sources
	    {  



                for(int i = 0; i < parameterCount("sourceno"); i++)
	        {
	            parameters.push_back(Parameters(sourceNo[i], photoType[i], 
                        radius[i], innerA[i], outerA[i], axialRatio[i], pa[i], background[i], weightingMethod) );
                }
            }
	    else // No fixed backgrounds
	    { 

              
            
                for(int i = 0; i < parameterCount("sourceno"); i++)
	        {
	            parameters.push_back(Parameters(sourceNo[i], photoType[i], 
                        radius[i], innerA[i], outerA[i], axialRatio[i], pa[i], 0.0, weightingMethod) );
                }
            }

          
	}        // Point source 
        else
	{
              
	
	    if( needBackground && needAperture && needBackgroundAnnulus )  // Using fixed background for all sources
	    {
         
                for(int i = 0; i < parameterCount("sourceno"); i++)
	        {
	            parameters.push_back(Parameters(sourceNo[i], photoType[i], radius[i], innerA[i], outerA[i], 0.0, 0.0, background[i], weightingMethod ));
                }

            }
            else if(needBackground && needAperture)
	    {

	      
                for(int i = 0; i < parameterCount("sourceno"); i++)
	        {
	            parameters.push_back(Parameters(sourceNo[i], photoType[i], radius[i], 0.0, 0.0, 0.0, 0.0, background[i], weightingMethod ));
                }


            }


	    else if( needAperture && needBackgroundAnnulus )  // Using fixed background for all sources
	    {
         
                for(int i = 0; i < parameterCount("sourceno"); i++)
	        {
	            parameters.push_back(Parameters(sourceNo[i], photoType[i], radius[i], innerA[i], outerA[i], 0.0, 0.0, 0.0, weightingMethod ));
                }

            }
  


            else if(needBackground)
	    {


        

                for(int i = 0; i < parameterCount("sourceno"); i++)
	        {
	            parameters.push_back(Parameters(sourceNo[i], photoType[i], 0.0, 0.0, 0.0, 0.0, 0.0, background[i], weightingMethod ));
                }

            }           
            else if(needAperture)
	    {

   

                for(int i = 0; i < parameterCount("sourceno"); i++)
	        {
	            parameters.push_back(Parameters(sourceNo[i], photoType[i], radius[i], 0.0, 0.0, 0.0, 0.0, 0.0, weightingMethod));
                }

            }
            else
	    {
               for(int i = 0; i < parameterCount("sourceno"); i++)
	        {  
		    

	            parameters.push_back(Parameters(sourceNo[i], photoType[i], 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, weightingMethod));
                }
            }
            



        }

        
        if(parameters.size() > 0)
	    process(parameters);
        else
            Sas::warning("No sources to reprocess!", "Please start again!");


    }


}
/******************************************************************************
*  Processes the sources for a given source-list, etc
*
******************************************************************************/

void process(vector<Parameters> &parameterList)
{

    // The input image
    string imageSet = stringParameter( "imageset" );


    // The input old source-list set
    string oldSrcListSet = stringParameter( "oldsrclistset" );

 

    // The new source-list set
    string newSrcListSet = stringParameter( "newsrclistset" );


    // The new source-list set
    string backgroundSet = stringParameter( "backgroundimageset" );


    



    if(backgroundSet.length() > 2)
    {

        if (!checkFiles(imageSet, backgroundSet, oldSrcListSet)) return;


        errstr << "Processing - imagefile=" << imageSet 
           << " backgroundimageset=" << backgroundSet
           << " oldsrclistset=" << oldSrcListSet
           << " newsrclistset=" << newSrcListSet 
  	   << Sas::message(Msg::App, Msg::Sparse);
           



    }
    else  
    {
        errstr << "Processing - imagefile=" << imageSet << " oldsrclistset=" << oldSrcListSet
           << " newsrclistset=" << newSrcListSet 
  	   << Sas::message(Msg::App, Msg::Sparse);

        if (!checkFiles(imageSet, oldSrcListSet)) return;

    }



    int binax1 = 0;
    

    getWindowData( imageSet, binax1);

    float minRad = float((12 / binax1) / 3);
    float maxRad = 12 / binax1;
    



    // The input old source-list set


   
  
    /***********************************************************
    * Load the sources from the old source-list file
    *************************************************************/
    int nSources = 0;
    F90MODPROC( omphotom_f1, initialise_data )( &nSources );
   
    if( nSources == 0)
    {
     
      Sas::fatal("No sources in source-list file", oldSrcListSet.c_str());
      

    }


    vector< OmSourceParameter > newSource(parameterList.size() );
    
    

    /********************************************************
    * Recompute photometry for each source in the list
    *********************************************************/

    vector<int> number; // Stores the number of the source that has been re-computed

   
    for(unsigned i = 0; i < parameterList.size(); i++)
    {

        int no = parameterList[i].sourceNo;

       
        if(no < 0 || no > nSources)
	{

	    errstr << "Source no " << no <<  " No such source! - ignoring"                     
  	           << Sas::message(Msg::App, Msg::Sparse);
	    continue;

        } 

        if( parameterList[i].type == 0 || parameterList[i].type == 1)
	{
        

	  //	  cout << "Line 647 aperure radius=" << parameterList[i].radius << endl;
	  //         cout << "Size parameter list=" << parameterList.size() << endl; 






            if( parameterList[i].radius < minRad || parameterList[i].radius > maxRad)
	    {

	      errstr << "Source no " << no <<  " Aperture radius=" << parameterList[i].radius 
		     << " outside limits (" << minRad << "-" << maxRad << ") ignoring source"                     
  	             << Sas::message(Msg::App, Msg::Sparse);



	      continue;
            }



        } 

  

	errstr << "Recomputing photometry for source " << no
  	   << Sas::message(Msg::App, Msg::Sparse);

     
        /*******************************************************************
	* Get the old source
        ********************************************************************/
        OmSourceParameter oldSource;
        

 


        F90MODPROC( omphotom_f1, getoldsource )( &no, &oldSource);
      
    
        /***********************************************************************
	 * Set the weighting method for the PSF fitting
	 **********************************************************************/
        setParameter("weighting", parameterList[i].weightingMethod);
        

  


	F90MODPROC( omphotom_f1, recompute_photometry )( 
                   &parameterList[i].noTerms, 
	           parameterList[i].fitNo, 
	           &parameterList[i].type,
	           &parameterList[i].radius, 
	           &parameterList[i].innerA, 
	           &parameterList[i].outerA, 
	           &parameterList[i].axialRatio, 
                   &parameterList[i].theta,
                   &parameterList[i].background);

     

        /**************************************************************************************************
	* Write out to a log file if required
        ***************************************************************************************************/     

     
     	output.precision(3);    
	errstr.precision(3);    

    
        output.setf(ios::showpoint);
        errstr.setf(ios::showpoint);

        switch(parameterList[i].type)     
	{
	    case 0: // Aperture photometry
	            
                errstr << no << ") aperture radius=" << parameterList[i].radius << " inner background annulus radius=" 
		       << parameterList[i].innerA << " outer=" <<  parameterList[i].outerA << " Aperture photometry" 
                       << Sas::message(Msg::App, Msg::Sparse);
                                       
		if(logFile)
		{      
                        
	             output << no << ") aperture radius=" << parameterList[i].radius << " inner background annulus radius=" 
			     << parameterList[i].innerA << " outer=" <<  parameterList[i].outerA << " Aperture photometry" << endl; 
                }
                break;

            case 1: // Aperture photometry
	            
		 errstr << no << ") aperture radius=" << parameterList[i].radius << " background counts per pixel=" 
                       << parameterList[i].background
		       << " Aperture photometry" 
                       << Sas::message(Msg::App, Msg::Sparse);
                                       
		if(logFile)
		{      
	            output << no << ") aperture radius=" << parameterList[i].radius << " background counts per pixel=" 
                       << parameterList[i].background
			   << " Aperture photometry\n"; 
                     
                                       
	             
                }
                break;




            case 2: // Least-squares fit to PSF, background from annulus
                               
                errstr  << no << ") inner background annulus radius=" 
			<< parameterList[i].innerA << " outer=" <<  parameterList[i].outerA 
                        << " Least-squares fit to PSF excluding constant term- background from annulus" 
                        << Sas::message(Msg::App, Msg::Sparse);
             


		 if(logFile)
                     output  << no << ") inner background annulus radius=" 
			     << parameterList[i].innerA << " outer=" <<  parameterList[i].outerA 
                             << " Least-squares fit to PSF excluding constant term- background from annulus" << endl; 
             
             
                 break;
                    
          
            case 3: // Least-squares fit to PSF for several sources, including background.
	                         
	       errstr << no << ") Least-squares fit to PSF for " << parameterList[i].noTerms << " sources, including constant term for background"  
		        << Sas::message(Msg::App, Msg::Sparse);
		 if(logFile)
                                    
		   output << no << ") Least-squares fit to PSF for " << parameterList[i].noTerms << " sources, including constant term for background" << endl;  
	
             
                 break;


             case 4: // Least-squares fit to PSF for several sources, excluding background.
	                         
	       errstr << no << ") Least-squares fit to PSF for " << parameterList[i].noTerms << " sources, excluding constant term for background"  
		        << Sas::message(Msg::App, Msg::Sparse);
		 if(logFile)
                                    
		   output << no << ") Least-squares fit to PSF for " << parameterList[i].noTerms << " sources, excluding constant term for background" << endl;  
	
             
                 break;


     
             case 5: // Least-squares fit to PSF, including background and position for one or more sources.
  

                 errstr << no << ") Least-squares fit to PSF including constant term for background and fitting to positions"  
		        << Sas::message(Msg::App, Msg::Sparse);

		 if(logFile)
                         
	             output << no << ") Least-squares fit to PSF including constant term for background and fitting to  positions" << endl; 
                            
                 break;




             case 6: // Least-squares fit to PSF, excluding background and position for one or more sources.
  

                 errstr << no << ") Least-squares fit to PSF excluding constant term for background, and fitting to positions"  
		        << Sas::message(Msg::App, Msg::Sparse);

		 if(logFile)
                         
	             output << no << ") Least-squares fit to PSF, excludind constant term for background, and fitting to positions" << endl; 
                            
                 break;

	     case -1:  // Extended-source photometry
                               
	         errstr << no << ") aperture semi-major axis=" << parameterList[i].radius << " inner background semi-major axis=" 
		        << parameterList[i].innerA << " outer=" <<  parameterList[i].outerA << " axial ratio=" 
                        <<  parameterList[i].axialRatio << " PA=" <<  parameterList[i].theta 
                        << " extended-source photometry\n" << Sas::message(Msg::App, Msg::Sparse);          
	  

  
	         if(logFile)
                                    
	             output << no << ") aperture semi-major axis=" << parameterList[i].radius << " inner background semi-major axis=" 
		            << parameterList[i].innerA << " outer=" <<  parameterList[i].outerA << " axial ratio=" 
                            <<  parameterList[i].axialRatio << " PA=" <<  parameterList[i].theta 
                            << " extended-source photometry\n" << Sas::message(Msg::App, Msg::Sparse);          
	  

                 break;

	     case -2:  // Extended-source photometry - fixed background
                               
	         errstr << no << ") aperture semi-major axis=" << parameterList[i].radius << " axial ratio=" 
                        <<  parameterList[i].axialRatio << " PA=" <<  parameterList[i].theta 
                        << " background counts per pixel=" << parameterList[i].background
                        << " extended-source photometry\n" << Sas::message(Msg::App, Msg::Sparse);          
	  

  
	         if(logFile)
                                    
	             output << no << ") aperture semi-major axis=" << parameterList[i].radius << " axial ratio=" 
                        <<  parameterList[i].axialRatio << " PA=" <<  parameterList[i].theta 
                        << " background counts per pixel=" << parameterList[i].background
                        << " extended-source photometry\n";          
	
	  





        }
       
	
    	errstr.precision(4);    



        /*******************************************************************
	* Get the new source
        ********************************************************************/
        
        
        OmSourceParameter newSource;
        
        F90MODPROC( omphotom_f1, getnewsource )( &no, &newSource);



        errstr << "Old magnitude=" << oldSource.magnitude 
               << " new magnitude=" << newSource.magnitude << Sas::message(Msg::App, Msg::Sparse); 
        if(output)
	{
	    output.precision(4);    
	    output << "Old magnitude=" << oldSource.magnitude 
                   << " new magnitude=" << newSource.magnitude << endl;
        }
       
   

        /********************************************************
	* Remember the source number that has changed
        *********************************************************/
        number.push_back( no );
        

   


    }

   
     

    setParameter("sourceno", number ); 

   

    /**************************************************************************************************
    * Create the new source-list file
    ***************************************************************************************************/



    F90MODPROC( omphotom_f1, finish )();



}
/******************************************************************************
*   
*
*******************************************************************************/
void getWindowData(string file, int &binax1)
{ 

  

    DataSet * set = dataSetServer->open(file, Dal::Read);
      
    if (!set)
    {
        errstr << "Accessing " << file << " ..."
  	       << Sas::message(Msg::App, Msg::Sparse);
    }
    

    binax1 = set->attribute("BINAX1")->asInt();
    binax1 = (int)pow( 2.0, binax1);
     
    

        
    bool binbpe = set->attribute("BINBPE")->asBool();
   
    if (binbpe) {
        binax1 *= 2;
	   
    }
   	  
  



    dataSetServer->close(set);

   


}
/******************************************************************************************
* Inputs a list of image and source-list files, etc,  to be
* reprocessed
* NB type                
*    -1      extended source photometry using given background image
*    -2      extended source photometry using fixed value for the background
*     0      aperture photometry using image or background image for background
*     1      aperture photometry using fixed value for the background
*     2      psf fit using background from annulus
*     3      psf fit to one or more sources - background included
*     4      psf fit to one or more sources using with fixed background
*     5      psf fit to one or more sources with background included and fittiing to positions
*     6      psf fit to one or more sources with background fixed and fittiing to positions

********************************************************************************************/
void readParameterFile(vector<string> &imageFile,
                       vector<string> &backgroundImageFile,
                       vector<string> &oldSrcListFile, 
                       vector<string> &newSrcListFile,
                       vector< vector<Parameters> > &parameterList)
{


 
    string file = stringParameter("paramfile");
    ifstream input(file.c_str());


    int no = 0;   





    /******************************************************************
    * Read the parameter file
    *****************************************************************/ 
    while(input && !input.eof())
    {
        //   stringstream ss (stringstream::in | stringstream::out);

        string buffer;
	

     
        if (!getNextLine(input, buffer, 256, false)) break;
       
     

	string file1 = getString(buffer, 5);




        int binax1 = 0;
    
      
        getWindowData( file1, binax1);

        float minRad = float((12 / binax1) / 3);
        float maxRad = 12 / binax1;
    

        if (!getNextLine(input, buffer, 256)) break;


       


          
	string file2 = getString(buffer, 5);
       
  
        if(!getNextLine(input, buffer, 256)) break;
           

	string file3 = getString(buffer, 5);

     
        int noOfRows = getNoRows(file3);

          

        if (!getNextLine(input, buffer, 256)) break;
  

	string file4 = getString(buffer, 5);

        /*****************************************************************************
	 * Get rid of blanks in the file names
	 *****************************************************************************/

        file1 = removeBlanks( file1 );
     
        file2 = removeBlanks( file2 );
     
        file3 = removeBlanks( file3 );
     
        file4 = removeBlanks( file4 );
     
//	cout << file1 << " length=" << file1.length() << " actual length=" << length(file1) << endl;
//	cout << file2 << " length=" << file2.length() << " actual length=" << length(file2) << endl;
//	cout << file3 << " length=" << file3.length() << " actual length=" << length(file3)  << endl;
//      cout << file4 << " length=" << file4.length() << " actual length=" << length(file4)  << endl;


        /*******************************************************
	* Get the number of sources to recompute for this image
        ********************************************************/
        if(!getNextLine(input, buffer, 10)) break;
             

        int size = getInteger(buffer);
        if(size <= 0)
	{
            Sas::fatal("Invalid no of sources", buffer);
        }

   
  
        parameterList.resize(no + 1);



        imageFile.push_back( file1 );
        backgroundImageFile.push_back( file2 );

        oldSrcListFile.push_back( file3 );
        newSrcListFile.push_back( file4 );

      
        /*********************************************************
	* Load the parameters for each source
	********************************************************/
  




        for(int i = 0; i < size; i++)
	{
	 
	  
  
	    if(!getNextLine(input, buffer, 100)) // was 256
	    { 
	   

	      exit(0);
                break;
            }

   
      

          
            stringstream ss (stringstream::in | stringstream::out);
          
	  
            ss << buffer;
           
            double ax = 0.0, th = 0.0; // Axial ratio and position angle for extended sources
         
           
	    int k = getInteger(ss, buffer);

            if(k <= 0 || k > noOfRows)
	    {
                Sas::fatal("Invalid source no", buffer);
            }


	    /*****************************************************
	    * Get the photometry type
	    *****************************************************/  
            int type = getInteger(ss, buffer);

	    

            if(type < -2 || type > 6 )
	    {
                Sas::fatal("Invalid photometry type", buffer);
              
                
            }

	  

            double r = 0.0, a = 0.0, b = 0.0, background = 0.0;


	    if(type == -2 || type == 1 || type == 4 || type == 6)  // Get the background counts per pixel 
	    {

	        /**********************************************************
		* get the aperture and annulii radii
		*********************************************************/

                background = getDouble(ss, buffer);


    




            }



      


     
	    if( type < 2)            // Need the aperture radius 
	    {








	        /**********************************************************
		* get the aperture and annulii radii
		*********************************************************/
                r = getDouble(ss, buffer);

	      

                if(type >= 0)
		{
                    if(r < minRad || r > maxRad)
		    {


		        char text[100];
                        sprintf(text, "Input value=%3.1f- Allowed values=%3.1f-%3.1f", r, minRad, maxRad); 
                        Sas::fatal("Invalid aperture radius!", text);
	             }


                }


            }


	  
   

            if(type == -1 || type == 0 || type == 2)  // Need the inner and outer background radii
	    {
           

                a = getDouble(ss, buffer);
               


                if(a < r) // Inner background annulus
	        {
                    Sas::fatal("Inner background radius < aperture radius", buffer);
                }
            

                b = getDouble(ss, buffer);
               
   	


                if(b < (a + 2.0))   // Outer background annulus
	        {
                    Sas::fatal("Outer background annulus < (inner + 2.0) annulus radius ", buffer);
                }
            }  
            
      

            if(type < 0)  // Extended sources- need the axial ratio and position angle
	    { 
                
	        /**********************************************************************
		* get the axial ration and position angle (for extended sources only)
		*********************************************************************/
          
                ax = getDouble(ss, buffer);
                if(ax < 1.0 || ax > 20.0)
	        {
                    Sas::fatal("Invalid axial ratio (has to be >=1 and <=20)", buffer);
                }
       



		th = getDouble(ss, buffer);
                
                if(th < 0.0 || th > 180.0)
	        {
                    Sas::fatal("Invalid position angle (has to be >=0 and <=180)", buffer);
                }

       


	    } 

	 

	    Parameters parameters(k, type, r, a, b, ax, th, background, 1);
            parameters.fitNo[0] = k;    




            if(type == 3 || type == 4 || type == 5 || type == 6)  // Least-squares fit to several sources, including background 
	    {
	        // Get the number of sources to be fitted
  
                int noToFit = getInteger(ss, buffer);
             
	      
	    

                if(noToFit < 0 || noToFit > 5) // Can only fit up to 5 sources
	        {
                    Sas::fatal("Number of sources to fit = 0 || > 5", buffer);
                }
                
                noToFit++;
                parameters.noTerms = noToFit;
               

                /***************************************************************************
		* Get the source numbers of the additional sources to be used in the fit
                ****************************************************************************/
                for(int i = 1; i < noToFit; i++)
		{
                    int j = getInteger(ss, buffer);
                    
                    
		 


                    
                    if(j < 1) // 
	            {
                        Sas::fatal("Source number < 1", buffer);
                    }


 
                    if(j > noOfRows) // 
	            {
                        Sas::fatal("No such source", buffer);
                    }






      
                    if(j == k) // 
	            {
                        Sas::fatal("Fitting to the same two sources!", buffer);
                    }




                    parameters.fitNo[i] = j;    



                    

                }
                
                /*****************************************************************
		* Get the weighting method (may not be specified)
                ******************************************************************/         
                int weightingMethod = 1;
                if(!ss.eof())
                {
       

                    weightingMethod = getInteger(ss, buffer);
                    if(weightingMethod < 1 || weightingMethod > 2)
		    {
                        Sas::fatal("Invalid PSF weighting method (allowed values = 1 or 2", buffer);
                    }

                }


	      

                parameters.weightingMethod = weightingMethod;

	




            }



            parameterList[no].push_back(parameters);
      


        
        }
        no++;
                

    }

       
      

    
}
/****************************************************
*
****************************************************/
int getInteger(stringstream &ss, string &buffer)
{
     


   
    if(ss.eof())
    {
        Sas::fatal("Error inputting integer data", buffer);


    }



    ss >> buffer;


    char *endPtr = 0;
    int no = strtol(buffer.c_str(), &endPtr, 10);


    if(strlen(endPtr) != 0)
    { 
        Sas::fatal("Error inputting integer data", buffer);
        return -1;
    }
    else 
        return no;    
}
/****************************************************
*
****************************************************/
double getDouble(stringstream &ss, string &buffer)
{


    if(ss.eof())
    {
        Sas::fatal("Error inputting double data", buffer);


    }



    
    ss >> buffer;

    char *endPtr = 0;
    double no = strtod(buffer.c_str(), &endPtr);


   
    if(strlen(endPtr) != 0) 
    {  
   

        Sas::fatal("Error inputting double data", buffer);
    
    
        return -1.0;
    }
    else
        return no;
}



/****************************************************
*
****************************************************/
int getInteger(string &buffer)
{
    

    char *endPtr = 0;
    int no = strtol(buffer.c_str(), &endPtr, 10);

    if(strlen(endPtr) > 0)
    { 
        if( endPtr[0] != ' ' && endPtr[0] != '\n')
	{
            Sas::fatal("Error inputting integer data", buffer);
            return -1;
        }
    }
     
    return no;    
}
/****************************************************
*
****************************************************/
double getDouble(string buffer)
{
  

    char *endPtr = 0;
    double no = strtod(buffer.c_str(), &endPtr);


    if(strlen(endPtr) > 0)
    { 
        if( endPtr[0] != ' ' && endPtr[0] != '\n')

        {  

            Sas::fatal("Error inputting double data", buffer);
            return -1.0;
        }
    }
   
    return no;

}
/**************************************************
*
**************************************************/

string getString(string &buffer, int minLength)
{

    stringstream ss(buffer);
          
    string s;
    
    ss >> s;

    if(ss.fail() )
    {   
        Sas::fatal("Error inputting string data", buffer);
           
    }

    if(length(s) < minLength)
    {

    
        Sas::fatal("Invalid name", s);
  

    }
    
    
    return s;

}


/***************************************************************
*
***************************************************************/

bool getNextLine(ifstream &input, string &buffer, int length, bool warn)
{   
    
    
  

    while(1) {
    
        buffer.clear();
        buffer.resize(length);


        input.getline((char *)buffer.c_str(), length, '\n'); 
        
	
	if( ::length(buffer) > 0) return true;
    	 
      
        if( input.fail())
	{

       
  
	

	    if(input.eof() )
	      if(warn)
	      {
	          Sas::warning("End of file reached!", "Please check your parameter file");
              } 
              input.clear();
              return false; 
        }

    }              
    return false;

}

/**************************************************
*
**************************************************/
int length(string & s)
{

 

    int size = 0;
    for (int i = 0; i < (int)s.length(); i++)
    {

      if( isgraph(s[i]) ) size++;
	    
    }

 

    return size;
}
/*********************************************************
* Removes all blanks from the given string
*********************************************************/
string removeBlanks( string & s)
{


    string newString = s;
    int size = 0;
    for (int i = 0; i < (int)s.length(); i++)
    {

      if( isgraph(s[i]) )
      { 
        newString[size] = s[i];
        size++;
      }    	    
    }

    newString.resize( size ); 

    return newString;

}
    	 
/**********************************************************
*
***********************************************************/
bool checkFiles(string &imageSet, string &backgroundSet, string &oldSrcListSet)
{


    DataSet * set = dataSetServer->open(imageSet, Dal::Read);
      

  
    if (!set)
    {
        errstr << "Accessing " << imageSet << " ..."
  	       << Sas::message(Msg::App, Msg::Sparse);
        return false;
    }
    

    string obsId = set->attribute("OBS_ID")->asString();
 
    string expId = set->attribute("EXP_ID")->asString();
    string filter = set->attribute("FILTER")->asString();
 
 
    dataSetServer->close(set);

   
    set = dataSetServer->open(backgroundSet, Dal::Read);

    if (!set)
    {
        errstr << "Accessing " << backgroundSet << " ..."
  	       << Sas::message(Msg::App, Msg::Sparse);
        return false;
    }
    



      

    if(obsId != set->attribute("OBS_ID")->asString())
    {

        errstr << "OBS_ID in file " << imageSet << " not the same as that in file " << backgroundSet
               << " ignoring files"  
  	       << Sas::message(Msg::App, Msg::Sparse);

        dataSetServer->close(set);

        return false;

    }



    if(filter != set->attribute("FILTER")->asString())
    {

        errstr << "FILTER in file " << imageSet << " not the same as that in file " << backgroundSet
               << " ignoring files"  
  	       << Sas::message(Msg::App, Msg::Sparse);

        dataSetServer->close(set);

        return false;

    }



    dataSetServer->close(set);


    set = dataSetServer->open(oldSrcListSet, Dal::Read);

    if (!set)
    {
        errstr << "Accessing " << oldSrcListSet << " ..."
  	       << Sas::message(Msg::App, Msg::Sparse);
        return false;
    }
    


    if(expId != set->attribute("EXP_ID")->asString())
    {

        errstr << "OBS_ID in file " << imageSet << " not the same as that in file " << oldSrcListSet
               << " ignoring files"  
  	       << Sas::message(Msg::App, Msg::Sparse);

        dataSetServer->close(set);

        return false;

    }

    dataSetServer->close(set);


    return true;

}

/**********************************************************
*
***********************************************************/
bool checkFiles(string &imageSet, string &oldSrcListSet)
{


    DataSet * set = dataSetServer->open(imageSet, Dal::Read);
      

  
    if (!set)
    {
        errstr << "Accessing " << imageSet << " ..."
  	       << Sas::message(Msg::App, Msg::Sparse);
        return false;
    }
    
 
    string expId = set->attribute("EXP_ID")->asString();
    
 
    dataSetServer->close(set);

  


    set = dataSetServer->open(oldSrcListSet, Dal::Read);

    if (!set)
    {
        errstr << "Accessing " << oldSrcListSet << " ..."
  	       << Sas::message(Msg::App, Msg::Sparse);
        return false;
    }
    



    if(expId != set->attribute("EXP_ID")->asString())
    {

        errstr << "OBS_ID in file " << imageSet << " not the same as that in file " << oldSrcListSet
               << " ignoring files"  
  	       << Sas::message(Msg::App, Msg::Sparse);

        dataSetServer->close(set);

        return false;

    }

    dataSetServer->close(set);



    return true;


}



/**********************************************************
*
***********************************************************/
int getNoRows(string &file)
{

    DataSet * set = dataSetServer->open(file, Dal::Read);
        
    if (!set)
    {
        errstr << "Accessing " << file << " ..."
  	       << Sas::message(Msg::App, Msg::Sparse);
        return false;
    }


    const Table *table = set->table("SRCLIST");
    int noRows = table->rows();  

    dataSetServer->close(set);

    return noRows;


}


